from abc import ABC, abstractmethod
from dataclasses import dataclass
from io import BytesIO

@dataclass
class File:
    name: str
    data: BytesIO

class ISaverAdapter(ABC):

    @abstractmethod
    def save(self, path: str, file: File) -> bool:
        ...



from typing import Union
from enma.application.core.utils.logger import logger
from enma.domain.entities.manga import MIME


def get_mime_safelly(mime: str) -> Union[MIME, None]:
    try:
        return MIME[mime]
    except ValueError:
        logger.error(f'Invalid MIME type: {mime}')
        return None


from abc import ABC, abstractmethod
from typing import Any, Union
from enma.domain.entities.author_page import AuthorPage

from enma.domain.entities.manga import Chapter, Manga, SymbolicLink
from enma.domain.entities.pagination import Pagination
from enma.domain.entities.search_result import SearchResult

class IMangaRepository(ABC):

    @abstractmethod
    def set_config(self, 
                   config: Any) -> None:
        ...

    @abstractmethod
    def get(self,
            identifier: str,
            with_symbolic_links: bool) -> Union[Manga, None]:
        ...
    
    @abstractmethod
    def search(self,
               query: str,
               page: int,
               **kwargs) -> SearchResult:
        ...

    @abstractmethod
    def paginate(self,
                 page: int) -> Pagination: ...
    
    @abstractmethod
    def random(self) -> Manga:
        ...

    @abstractmethod
    def author_page(self,
                    author: str,
                    page: int) -> AuthorPage:
        ...

    @abstractmethod
    def fetch_chapter_by_symbolic_link(self,
                                       link: SymbolicLink) -> Chapter:
        ...



"""
This module provides an adapter for the MANGANATO repository.
It contains functions and classes to interact with the MANGANATO API and retrieve manga data.
"""

from concurrent.futures import ThreadPoolExecutor
from datetime import datetime
from multiprocessing import cpu_count
import os
from typing import Any, Optional, Union, cast
from urllib.parse import urlparse, urljoin
from bs4 import BeautifulSoup, Tag

import requests

from enma._version import __version__
from enma.application.core.interfaces.manga_repository import IMangaRepository
from enma.application.core.utils.logger import logger
from enma.domain.entities.author_page import AuthorPage
from enma.domain.entities.manga import Author, Chapter, Genre, Image, Manga, SymbolicLink, Title
from enma.domain.entities.search_result import Pagination, SearchResult, Thumb
from enma.infra.core.utils.cache import Cache

class Manganato(IMangaRepository):
    """
    Repository class for interacting with the Manganato website.
    Provides methods to fetch manga details, search for manga, etc.
    """
    def __init__(self) -> None:
        self.__BASE_URL = 'https://manganato.com'
        self.__CHAPTER_BASE_URL = 'https://chapmanganato.com'

    def __make_request(self, 
                       url: str,
                       headers: Union[dict[str, Any], None] = None,
                       params: Optional[dict[str, Union[str, int]]] = None):

        headers = headers if headers is not None else {}
        params = params if params is not None else {}

        logger.debug(f'Fetching {url} with headers {headers} and params {params}')

        return requests.get(url=urlparse(url).geturl(), 
                            headers={**headers, 
                                     'Referer': 'https://chapmanganato.com/', 
                                     "User-Agent": f"Enma/{__version__}"},
                            params={**params})
    
    def __create_title(self, 
                       main_title: str, 
                       alternative: str) -> Title:
        logger.debug(f'Building manga title main: {main_title} and alternative: {alternative}')

        has_many_alternatives = alternative.find(';') != -1 or alternative.find(',') != -1

        if not has_many_alternatives:
            jp = alternative
            return Title(english=main_title.strip(),
                         japanese=jp.strip(),
                         other=main_title.strip())

        jp, cn, *_ = alternative.split(';') if alternative.find(';') != -1 else alternative.split(',')
        return Title(english=main_title.strip(),
                     japanese=jp.strip(),
                     other=cn.strip())
    
    def __find_chapters_list(self, html: BeautifulSoup) -> list[str]:
        chapter_list = cast(Tag, html.find('ul', {'class': 'row-content-chapter'}))
        chapters = chapter_list.find_all('li') if chapter_list else []
        return [chapter.find('a')['href'] for chapter in chapters]
    
    def __create_chapter(self, url: str, symbolic: bool = False) -> Union[Chapter, None]:

        if symbolic:
            return Chapter(id=url.split('/')[-1], link=SymbolicLink(link=url))

        response = self.__make_request(url=url)
        logger.debug(f'Fetching chapter {url}')
        
        if response.status_code != 200:
            logger.error(f'Could not fetch the chapter with url: {url}. status code: {response.status_code}')
            return
        
        chapter = Chapter(id=response.url.split('/')[-1])
        html = BeautifulSoup(response.text, 'html.parser')
        images_container = cast(Tag, html.find('div', {'class': 'container-chapter-reader'}))

        for img in images_container.find_all('img'):
            chapter.add_page(Image(uri=img['src'],
                                   name=img['src'].split('/')[-1],
                                   width=0,
                                   height=0))
        return chapter
    
    def set_config(self, **kwargs) -> None:
        raise NotImplementedError('Manganato does not support set_config')

    @Cache(max_age_seconds=int(os.getenv('ENMA_CACHING_GET_TTL_IN_SECONDS', 300)), 
           max_size=20).cache
    def get(self, 
            identifier: str,
            with_symbolic_links: bool = False) -> Union[Manga, None]:
        response = self.__make_request(url=urljoin(self.__CHAPTER_BASE_URL, identifier))
        
        if response.status_code != 200:
            logger.error(f'Could not fetch the manga with identifier: {identifier}. status code: {response.status_code}')
            return
        
        soup = BeautifulSoup(response.text, 'html.parser')

        title: Title = Title()
        author: Optional[str] = None
        genres: list[str] = []
        cover = None

        elem_title = ''

        cover_img = cast(Tag, soup.find('span', {'class': 'info-image'}))
        if cover_img: 
            cover_tag = cast(Tag, cover_img.find('img'))
            if cover_tag:
                cover = cover_tag['src']

        title_div = soup.find('div', {'class': 'story-info-right'})

        if title_div is None: return

        h1_elem = title_div.find('h1')
        h1_tag = cast(Tag, h1_elem)
        h1_text = h1_tag.text if h1_elem is not None else ''
        elem_title = h1_text
    
        table = cast(Tag, soup.find('table', {'class': 'variations-tableInfo'}))
        
        if table is None: return

        table_vals = table.find_all('td', {'class': 'table-value'})
        
        if table_vals is None: return
        
        title_cel, author_cel, status_cel, genres_cel = table_vals
        title = self.__create_title(main_title=elem_title, 
                                    alternative=title_cel.text)
        author = author_cel.text.strip()
        status = status_cel.text.strip().lower()
  
        genres = genres_cel.text.replace('\n', '').split(' - ')

        extra_infos = cast(Tag, soup.find('div', {'class': 'story-info-right-extent'}))

        updated_at = None
        if extra_infos:
            updated_at_field = extra_infos.find_all('p')[0]
            updated_at = updated_at_field.find('span', {'class': 'stre-value'}).text

        if with_symbolic_links:
            chapters_links = self.__find_chapters_list(html=soup)
            chapters = [self.__create_chapter(link, symbolic=True) for link in chapters_links]
        else:
            workers = cpu_count()
            logger.debug(f'Initializing {workers} workers to fetch chapters of {identifier}.')

            with ThreadPoolExecutor(max_workers=workers) as executor:
                chapters = executor.map(self.__create_chapter, self.__find_chapters_list(html=soup))
                chapters = list(filter(lambda x: isinstance(x, Chapter), list(chapters)))
                executor.shutdown()
        
        return Manga(title=title,
                     status='completed' if status == 'completed' else 'ongoing',
                     authors=[Author(name=author)] if author is not None else None,
                     genres=[Genre(name=genre_name) for genre_name in genres],
                     url=urljoin(self.__BASE_URL, identifier),
                     id=identifier,
                     created_at=datetime.strptime(updated_at, "%b %d,%Y - %H:%M %p") if updated_at else None,
                     updated_at=datetime.strptime(updated_at, "%b %d,%Y - %H:%M %p") if updated_at else None,
                     thumbnail=Image(uri=cover), # type: ignore
                     cover=Image(uri=cover), # type: ignore
                     chapters=chapters) # type: ignore
    
    @Cache(max_age_seconds=int(os.getenv('ENMA_CACHING_SEARCH_TTL_IN_SECONDS', 100)), 
           max_size=5).cache
    def search(self, 
               query: str,
               page: int) -> SearchResult:
        
        response = self.__make_request(url=urljoin(f'{self.__BASE_URL}/search/story/', "_".join(query.split())),
                                       params={'page': page})
        
        if response.status_code != 200:
            return SearchResult(query=query,
                                page=page)
        
        soup = BeautifulSoup(response.text, 'html.parser')

        total_pages = 0
        total_pages_elem = soup.find('a', {'class': 'page-blue page-last'})
        if total_pages_elem:
            total_pages = int(total_pages_elem['href'].split('page=')[-1]) # type: ignore

        results = soup.find_all('div', {'class': 'search-story-item'})

        thumbs: list[Thumb] = []

        for result in results:
            thumbs.append(Thumb(title=result.find('h3').text.replace('\n', '').strip(),
                                url=result.find('a', {'class': 'a-h text-nowrap item-title'})['href'],
                                cover=Image(uri=result.find('img')['src'], width=0, height=0),
                                id=result.find('a', {'class': 'a-h text-nowrap item-title'})['href'].split('/')[-1]))

        return SearchResult(query=query,
                            page=page,
                            total_pages=total_pages,
                            results=thumbs)

    @Cache(max_age_seconds=int(os.getenv('ENMA_CACHING_PAGINATE_TTL_IN_SECONDS', 100)), 
           max_size=5).cache
    def paginate(self, page: int) -> Pagination:
        response = self.__make_request(url=f'{self.__BASE_URL}/genre-all/{page}')
        
        pagination = Pagination(page=page)

        if response.status_code != 200:
            return pagination
        
        soup = BeautifulSoup(response.text, 'html.parser')

        content_panel = soup.find('div', {'class': 'panel-content-genres'})
        
        if content_panel is None:
            return pagination
        
        content_panel = cast(Tag, content_panel)
        content_items = content_panel.find_all('div', {'class': 'content-genres-item'})
        
        for item in content_items:
            item = cast(Tag, item)
            info = cast(Tag, item.find('a', {'class': 'genres-item-img bookmark_check'}))
            cover = info.find('img')
            pagination.results.append(Thumb(id=info['href'].split('/')[-1], # type: ignore
                                            url=info['href'], # type: ignore
                                            title=info['title'] if info is not None else "", # type: ignore
                                            cover=Image(uri=cover['src'], width=0, height=0))) # type: ignore
            
        last_pagination = soup.find('a', {'class': 'page-blue page-last'})
        pagination.total_pages = int(last_pagination['href'].split('/')[-1]) # type: ignore
        pagination.total_results = 24 * int(pagination.total_pages)
        
        return pagination
    
    def random(self) -> Manga:
        raise NotImplementedError('Manganato does not support random')
    
    def author_page(self, author: str, page: int) -> AuthorPage:
        raise NotImplementedError('Manganato does not support author_page')
    
    @Cache(max_age_seconds=int(os.getenv('ENMA_CACHING_FETCH_SYMBOLIC_LINK_TTL_IN_SECONDS', 100)), 
           max_size=20).cache
    def fetch_chapter_by_symbolic_link(self, link: SymbolicLink) -> Chapter:
        chapter =  self.__create_chapter(url=link.link)
        
        if chapter is not None:
            return chapter
        
        return Chapter(id=0)



from datetime import datetime
from typing import Generic, TypeVar, Union
from uuid import uuid4

T = TypeVar('T')

class Entity(Generic[T]):
    """Base class for entities in the domain model.

    This class provides common attributes and methods for all entities.

    Attributes:
        id: A Union of int and str representing the entity's ID.
        created_at: A datetime object representings when the entity was created.
        updated_at: A datetime object representing when the entity was last updated.
    """

    def __init__(self, 
                 id: Union[int, str, None] = None,
                 created_at: Union[datetime, None] = None,
                 updated_at: Union[datetime, None] = None) -> None:
        """Initializes an Entity with given or default values.

        Args:
            id: A Union of int, str, and None representing the entity's ID. Defaults to uuidv4.
            created_at: A Union of datetime and None representing when the entity was created. Defaults to current UTC time.
            updated_at: A Union of datetime and None representing when the entity was last updated. Defaults to current UTC time.
        """

        self.id = id if id is not None else str(uuid4())
        self.created_at = created_at if created_at is not None else datetime.utcnow()
        self.updated_at = updated_at if updated_at is not None else datetime.utcnow()

    def __repr__(self) -> str:
        non_special_attrs = [f"{chave}={valor!r}" for chave, valor in self.__dict__.items() if not isinstance(valor, list) and not isinstance(valor, dict)]
        special_attrs = [f"{chave}={valor!r}" for chave, valor in self.__dict__.items() if isinstance(valor, list) or isinstance(valor, dict)]
        return f"{self.__class__.__name__}({', '.join([*non_special_attrs, *special_attrs])})"

    def to_dict(self) -> T:
        """Converts the entity to a dictionary.

        Returns:
            A dictionary representation of the entity.
        """
        return self.__dict__ # type: ignore



"""
This module defines the core interfaces for the Enma application's infrastructure layer.
It provides abstract classes and methods that need to be implemented by the concrete classes.
"""
from abc import ABC, abstractmethod
from enma.application.core.interfaces.downloader_adapter import IDownloaderAdapter
from enma.application.core.interfaces.saver_adapter import ISaverAdapter
from enma.application.use_cases.download_chapter import Threaded
from enma.domain.entities.author_page import AuthorPage

from enma.domain.entities.manga import Chapter, Manga, SymbolicLink
from enma.domain.entities.pagination import Pagination
from enma.domain.entities.search_result import SearchResult

class IEnma(ABC):
    @abstractmethod
    def get(self, 
            identifier: str,
            with_symbolic_links: bool) -> Manga:
        ...
    
    @abstractmethod
    def search(self, query: str, page: int, **kwargs) -> SearchResult:
        ...
    
    @abstractmethod
    def paginate(self, page: int) -> Pagination:
        ...

    @abstractmethod
    def random(self) -> Manga:
        ...

    @abstractmethod
    def download_chapter(self,
                         path: str,
                         chapter: Chapter,
                         downloader: IDownloaderAdapter,
                         saver: ISaverAdapter,
                         threaded: Threaded) -> None:
        ...
    
    @abstractmethod
    def author_page(self,
                    author: str,
                    page: int) -> AuthorPage: 
        ...
    
    @abstractmethod
    def fetch_chapter_by_symbolic_link(self,
                                       link: SymbolicLink) -> Chapter:
        ...



"""
This module provides an adapter for the nhentai repository.
It contains functions and classes to interact with the nhentai API and retrieve manga data.
"""
from datetime import datetime, timezone
from enum import Enum
import os
from typing import Any, List, Literal, Optional, Union, cast
from urllib.parse import urljoin, urlparse
from pydantic import BaseModel, field_validator

import requests
from bs4 import BeautifulSoup, Tag

from enma.application.core.handlers.error import (ExceedRetryCount, Forbidden, InvalidConfig, InvalidRequest,
                                                  NotFound, Unknown)
from enma.application.core.interfaces.manga_repository import IMangaRepository
from enma.application.core.utils.logger import logger
from enma.domain.entities.author_page import AuthorPage
from enma.domain.entities.manga import (MIME, Chapter, Genre, Author, Image, Manga, SymbolicLink,
                                        Title, Tag as EnmaTag)
from enma.domain.entities.search_result import Pagination, SearchResult, Thumb
from enma.domain.utils import mime
from enma.infra.core.interfaces.nhentai_response import NHentaiImage, NHentaiResponse, Tag as NHentaiResponseTag
from enma.infra.core.utils.cache import Cache
from enma._version import __version__


class CloudFlareConfig(BaseModel):
    user_agent: str
    cf_clearance: str

    @field_validator('user_agent')
    def user_agent_validator(cls, value: str) -> str:
        if value == '': raise InvalidRequest(message='User Agent cant be empty.')
        return str(value)

    @field_validator('cf_clearance')
    def cf_clearance_validator(cls, value: str) -> str:
        if value == '': raise InvalidRequest(message='CF Clearance cant be empty.')
        return str(value)

class __StrEnum(str, Enum):...

class Sort(__StrEnum):
    TODAY = 'popular-today'
    WEEK = 'popular-week'
    ALL_TIME = 'popular'
    RECENT = None

TagType = Union[Literal["language"],
                Literal["parody"],
                Literal["character"],
                Literal["group"],
                Literal["artist"],
                Literal["tag"],
                Literal["category"]]

class NHentai(IMangaRepository):
    """
    Repository class for interacting with the nhentai API.
    Provides methods to fetch manga details, search for manga, etc.
    """
    def __init__(self,
                 config: Optional[CloudFlareConfig] = None) -> None:
        self.__config = config
        self.__BASE_URL = 'https://nhentai.net/'
        self.__API_URL = 'https://nhentai.net/api/'
        self.__IMAGE_BASE_URL = 'https://i.nhentai.net/galleries/'
        self.__AVATAR_URL = 'https://i5.nhentai.net/'
        self.__TINY_IMAGE_BASE_URL = self.__IMAGE_BASE_URL.replace('/i.', '/t.')

    def __handle_source_response(self, response: requests.Response):
        logger.debug(f'Fetched {response.url} with response status code {response.status_code} and text {response.text}')
        if response.status_code == 200: return
        if response.status_code == 403: 
            raise Forbidden(message='Could not perform a successfull request to the source due credentials issues. \
Check your credentials and try again.')
        if response.status_code == 404:
            raise NotFound(message=f'Could not find the requested resource at "{response.url}". \
Check the provided request parameters and try again.')
        raise Unknown(message='Something unexpected happened while trying to fetch source content. \
Set the logging mode to debug and try again.')

    def __make_request(self,
                       url: str,
                       headers: Union[dict[str, Any], None] = None,
                       params: Optional[dict[str, Union[str, int]]] = None) -> requests.Response:

        headers = headers if headers is not None else {}
        params = params if params is not None else {}

        config = self.__config
        user_agent = config.user_agent if config is not None else f"Enma/{__version__}"
        cookies = {'cf_clearance': config.cf_clearance} if config is not None else {}

        logger.debug(f'Fetching {url} with headers {headers}, params {params} and cookies: {cookies}')

        response = requests.get(url=urlparse(url).geturl(),
                                headers={**headers, 'User-Agent': user_agent},
                                params={**params},
                                cookies=cookies)
        
        self.__handle_source_response(response)

        return response

    def set_config(self, config: CloudFlareConfig) -> None:
        if not isinstance(config, CloudFlareConfig): raise InvalidConfig(message='You must provide a CloudFlareConfig object.') 
        self.__config = config
    
    def __make_page_uri(self,
                        type: Union[Literal['cover'], Literal['page'], Literal['thumbnail']],
                        media_id: str,
                        mime: MIME,
                        page_number: Optional[int] = None) -> str:
        
        url = ''

        # cannot use switch case because this lib intends to support python 3.9+
        if type == 'cover': 
            url = urljoin(self.__TINY_IMAGE_BASE_URL, f'{media_id}/cover.{mime.value}')
        elif type == 'thumbnail': 
            url = urljoin(self.__TINY_IMAGE_BASE_URL, f'{media_id}/thumb.{mime.value}')
        else:
            url = urljoin(self.__IMAGE_BASE_URL, f'{media_id}/{page_number}.{mime.value}')
        
        logger.debug(f'Built page uri for type {type} as {url}')

        return url
    
    @Cache(max_age_seconds=int(os.getenv('ENMA_CACHING_FETCH_SYMBOLIC_LINK_TTL_IN_SECONDS', 100)), 
           max_size=20).cache
    def fetch_chapter_by_symbolic_link(self, 
                                       link: SymbolicLink) -> Chapter:
        response = self.__make_request(url=link.link)
        
        doujin: NHentaiResponse = response.json()

        if doujin.get('media_id') is None or doujin.get('images') is None:
            return Chapter()

        return self.__create_chapter(url=link.link, 
                                     with_symbolic_links=False, 
                                     media_id=doujin.get('media_id'), 
                                     pages=doujin.get('images').get('pages'))

    def __create_chapter(self,
                         url: str,
                         media_id: str, 
                         pages: list[NHentaiImage],
                         with_symbolic_links: bool = False) -> Chapter:
        
        if with_symbolic_links:
            return Chapter(link=SymbolicLink(link=url))
        else:
            chapter = Chapter()
            for index, page in enumerate(pages):
                safe_mime = mime.get_mime_safelly(page.get('t').upper())

                if safe_mime is None:
                    logger.warning(f'Could not find a valid mime type for page {index+1}. Forcing mime type as JPG.')

                safe_mime = safe_mime if safe_mime is not None else MIME.J

                chapter.add_page(Image(uri=self.__make_page_uri(type='page',
                                                                mime=safe_mime,
                                                                media_id=media_id,
                                                                page_number=index+1),
                                name=f'{index}.{safe_mime.value}',
                                mime=safe_mime,
                                width=page.get('w'),
                                height=page.get('h')))
            return chapter
        
    def __get_tag_by_type(self, 
                          type: TagType, 
                          tags: List[NHentaiResponseTag]) -> List[NHentaiResponseTag]:
        return list(filter(lambda tag: tag.get('type') == type, tags))
    
    @Cache(max_age_seconds=int(os.getenv('ENMA_CACHING_GET_TTL_IN_SECONDS', 300)), 
           max_size=20).cache
    def get(self, 
            identifier: str,
            with_symbolic_links: bool = False) -> Union[Manga, None]:

        url = urljoin(self.__API_URL, f'gallery/{identifier}')
        response = self.__make_request(url=url)

        doujin: NHentaiResponse = response.json()
        media_id = doujin.get('media_id')

        nhentai_tags = doujin.get('tags')

        chapter = self.__create_chapter(url=url,
                                        with_symbolic_links=with_symbolic_links,
                                        media_id=media_id, 
                                        pages=doujin.get('images').get('pages'))

        language = [tag.get('name') 
                    for tag in self.__get_tag_by_type(type='language',
                                                      tags=nhentai_tags)]
        
        authors = [Author(id=tag.get('id'),
                          name=tag.get('name')) 
                    for tag in self.__get_tag_by_type(type='artist',
                                                      tags=nhentai_tags)]
        
        genres = [Genre(id=tag.get('id'),
                        name=tag.get('name')) 
                    for tag in self.__get_tag_by_type(type='tag',
                                                      tags=nhentai_tags)]
        
        characters = [EnmaTag(type='character',
                              name=tag.get('name'),
                              id=tag.get('id')) 
                      for tag in self.__get_tag_by_type(type='character',
                                                        tags=nhentai_tags)]
        
        related = [EnmaTag(type='related',
                           name=tag.get('name'),
                           id=tag.get('id')) 
                      for tag in self.__get_tag_by_type(type='parody',
                                                        tags=nhentai_tags)]
        
        category = [EnmaTag(type='category',
                            name=tag.get('name'),
                            id=tag.get('id')) 
                      for tag in self.__get_tag_by_type(type='category',
                                                        tags=nhentai_tags)]
        
        tags = [*characters, *related, *category]
        
        safe_mime = mime.get_mime_safelly(doujin.get('images').get('thumbnail').get('t').upper())
        safe_mime = safe_mime if safe_mime is not None else MIME.J

        thumbnail = Image(uri=self.__make_page_uri(type='thumbnail',
                                                   mime=safe_mime,
                                                   media_id=media_id),
                          mime=safe_mime,
                          width=doujin.get("images").get("thumbnail").get("w"),
                          height=doujin.get("images").get("thumbnail").get("h"))
        
        safe_mime = mime.get_mime_safelly(doujin.get("images").get("cover").get("t").upper())
        safe_mime = safe_mime if safe_mime is not None else MIME.J

        cover = Image(uri=self.__make_page_uri(type='cover',
                                               media_id=media_id,
                                               mime=safe_mime),
                      mime=safe_mime,
                      width=doujin.get("images").get("cover").get("w"),
                      height=doujin.get("images").get("cover").get("h"))

        manga = Manga(title=Title(english=doujin.get('title').get('english'),
                                  japanese=doujin.get('title').get('japanese'),
                                  other=doujin.get('title').get('pretty')),
                      id=doujin.get('id'),
                      created_at=datetime.fromtimestamp(doujin.get('upload_date'), tz=timezone.utc),
                      updated_at=datetime.fromtimestamp(doujin.get('upload_date'), tz=timezone.utc),
                      status='completed',
                      url=urljoin(self.__BASE_URL, f'g/{doujin.get("id")}'),
                      language=language[0] if len(language) > 0 else None,
                      authors=authors,
                      genres=genres,
                      thumbnail=thumbnail,
                      tags=tags,
                      cover=cover,
                      chapters=[chapter])

        return manga
    
    @Cache(max_age_seconds=int(os.getenv('ENMA_CACHING_SEARCH_TTL_IN_SECONDS', 100)), 
           max_size=5).cache
    def search(self,
               query: str,
               page: int,
               sort: Sort = Sort.RECENT) -> SearchResult:
        
        logger.debug(f'Searching into Nhentai with args query={query};page={page};sort={sort}')

        request_response = self.__make_request(url=urljoin(self.__BASE_URL, 'search'),
                                               params={'q': query,
                                                       'sort': sort if isinstance(sort, str) else sort.value,
                                                       'page': page})
        
        search_result = SearchResult(query=query,
                                     total_pages=0,
                                     page=page,
                                     total_results=0,
                                     results=[])
        
        soup = BeautifulSoup(request_response.text, 'html.parser')

        search_results_container = soup.find('div', {'class': 'container'})
        logger.debug(f'Found successfully search result container using .class.container.')
        pagination_container = soup.find('section', {'class': 'pagination'})
        logger.debug(f'Found successfully pagination container using .class.pagination.')

        last_page_a_tag = pagination_container.find('a', {'class': 'last'}) if pagination_container else None # type: ignore
        logger.debug(f'Found last pagination container using .calss.last.')
        total_pages = int(last_page_a_tag['href'].split('=')[-1]) if last_page_a_tag else 1 # type: ignore
        logger.debug(f'Found last page number successfully splitting last pagination number href.')

        if not search_results_container:
            return search_result

        search_results = search_results_container.find_all('div', {'class': 'gallery'}) # type: ignore
        logger.debug(f'Found search results container using .class.gallery')
        
        if not search_results:
            return search_result
        
        a_tags_with_doujin_id = [gallery.find('a', {'class': 'cover'}) for gallery in search_results]

        thumbs = []

        for a_tag in a_tags_with_doujin_id:
            if a_tag is None: continue

            doujin_id = a_tag['href'].split('/')[-2]

            if doujin_id == '': continue

            result_cover = a_tag.find('img', {'class': 'lazyload'})
            cover_uri = None
            width = None
            height = None

            if result_cover is not None:
                cover_uri = result_cover['data-src']
                width = result_cover['width']
                height = result_cover['height']

            result_caption = a_tag.find('div', {'class': 'caption'})

            caption = None
            if result_caption is not None:
                caption = result_caption.text

            thumbs.append(Thumb(id=doujin_id,
                                url=urljoin(self.__BASE_URL, f'g/{doujin_id}'),
                                cover=Image(uri=cover_uri or '',
                                            mime=MIME.J,
                                            width=int(width or 0),
                                            height=int(height or 0)),
                                title=caption or ''))

        search_result.total_pages = total_pages
        search_result.total_results = 25 * total_pages if pagination_container else len(thumbs)
        search_result.results = thumbs

        return search_result

    @Cache(max_age_seconds=int(os.getenv('ENMA_CACHING_PAGINATE_TTL_IN_SECONDS', 100)), 
           max_size=5).cache
    def paginate(self, page: int) -> Pagination:
        response = self.__make_request(url=urljoin(self.__API_URL, f'galleries/all'),
                                       params={'page': page})

        data = response.json()

        PAGES = data.get('num_pages', 0)
        PER_PAGE = data.get('per_page', 0)
        TOTAL_RESULTS = int(PAGES) * int(PER_PAGE)


        
        pagination = Pagination(page=int(page),
                                total_results=TOTAL_RESULTS,
                                total_pages=PAGES,
                                results=[])
        
        for result in data.get('result'):
            safe_mime = mime.get_mime_safelly(result.get('images').get('cover').get('t').upper())
            safe_mime = safe_mime if safe_mime is not None else MIME.J
            thumb = Thumb(
                id=result.get('id'),
                title=result.get('title').get('english'),
                url=urljoin(self.__BASE_URL, f'g/{result.get("id")}'),
                cover=Image(
                    uri=self.__make_page_uri(
                        type='cover',
                        media_id=result.get('media_id'),
                        mime=safe_mime
                    ),
                    mime=safe_mime,
                    width=result.get('images').get('cover').get('w'),
                    height=result.get('images').get('cover').get('h')
                )
            )

            pagination.add_result(thumb)

        return pagination

    def random(self, retry=0) -> Manga:
        response = self.__make_request(url=urljoin(self.__BASE_URL, 'random'))

        soup = BeautifulSoup(response.text, 'html.parser')

        id = cast(Tag, soup.find('h3', id='gallery_id')).text.replace('#', '')
        logger.debug(f'Got successfully random manga id {id} from "gallery_id.h3" tag.')
        logger.debug(f'Using get method to Fetch manga with identifier: {id}')
        doujin = self.get(identifier=id)
        logger.debug(f'Got successfully random manga with id {id}.')

        if doujin is None:
            if retry >= 4:
                raise ExceedRetryCount('Could not fetch a random doujin after 4 retries.')
            
            logger.warning('Could not fetch random manga. Retrying.')
            return self.random(retry=retry+1)

        return doujin
    
    @Cache(max_age_seconds=int(os.getenv('ENMA_CACHING_AUTHOR_TTL_IN_SECONDS', 100)), 
           max_size=5).cache
    def author_page(self,
                    author: str,
                    page: int) -> AuthorPage:
        request_response = self.__make_request(url=urljoin(self.__BASE_URL, f'artist/{author}'),
                                               params={'page': page})
        
        result = AuthorPage(author=author,
                            total_pages=0,
                            page=page,
                            total_results=0,
                            results=[]) 
   
        soup = BeautifulSoup(request_response.text, 'html.parser')

        search_results_container = soup.find('div', {'class': 'container'})
        pagination_container = soup.find('section', {'class': 'pagination'})

        last_page_a_tag = pagination_container.find('a',
                                                    {'class': 'last'}) if pagination_container else None  # type: ignore
        total_pages = int(last_page_a_tag['href'].split('=')[-1]) if last_page_a_tag else 1  # type: ignore

        if not search_results_container:
            return result

        search_results = search_results_container.find_all('div', {'class': 'gallery'})  # type: ignore

        if not search_results:
            return result

        a_tags_with_doujin_id = [gallery.find('a', {'class': 'cover'}) for gallery in search_results]

        thumbs = []

        for a_tag in a_tags_with_doujin_id:
            if a_tag is None: continue

            doujin_id = a_tag['href'].split('/')[-2]

            if doujin_id == '': continue

            result_cover = a_tag.find('img', {'class': 'lazyload'})
            cover_uri = None
            width = None
            height = None

            if result_cover is not None:
                cover_uri = result_cover['data-src']
                width = result_cover['width']
                height = result_cover['height']

            result_caption = a_tag.find('div', {'class': 'caption'})

            caption = None
            if result_caption is not None:
                caption = result_caption.text

            thumbs.append(Thumb(id=doujin_id,
                                url=urljoin(self.__BASE_URL, f'g/{doujin_id}'),
                                cover=Image(uri=cover_uri or '',
                                            mime=MIME.J,
                                            width=int(width or 0),
                                            height=int(height or 0)),
                                title=caption or ''))
        
        result.author = author
        result.total_pages = total_pages
        result.page = page
        result.total_results = 25 * total_pages if pagination_container else len(thumbs)
        result.results = thumbs
        
        return result



from concurrent.futures import ThreadPoolExecutor, as_completed
from dataclasses import dataclass, field
import queue
import threading
from typing import Optional
from enma.application.core.interfaces.downloader_adapter import IDownloaderAdapter
from enma.application.core.interfaces.saver_adapter import File, ISaverAdapter
from enma.application.core.interfaces.use_case import DTO, IUseCase
from enma.application.core.utils.logger import logger
from enma.domain.entities.manga import Chapter, Image

@dataclass
class Threaded:
    use_threads: bool
    number_of_threads: int

@dataclass
class DownloadChapterRequestDTO:
    chapter: Chapter
    path: str
    saver_adapter: ISaverAdapter
    downloader: IDownloaderAdapter
    threaded: Optional[Threaded] = field(default_factory=lambda: Threaded(use_threads=False,
                                                                          number_of_threads=0))

@dataclass
class DownloadChapterResponseDTO:
    done: bool
    
class DownloadChapterUseCase(IUseCase[DownloadChapterRequestDTO, DownloadChapterResponseDTO]):

    def __make_download_with_queue(self, page: Image):
        downloaded = self.downloader.download(page=page)
        logger.debug(f'Download completed sending {page.name} to saving queue.')
        self.queue.put(File(name=page.name, data=downloaded))

    def __handle_saving(self, save_queue: queue.Queue[File], path: str, saver: ISaverAdapter) -> None:
        while True:
            try:
                content = save_queue.get(timeout=5)
                logger.debug(f'Saving file to path: {path} with name: {content.name}')
                saved = saver.save(path=path, file=content)
                if saved:
                    save_queue.task_done()
            except queue.Empty:
                return

    def __spawn_workers(self, chapter: Chapter, workers: int) -> None:
        with ThreadPoolExecutor(max_workers=workers) as executor:
            futures = {executor.submit(self.__make_download_with_queue, page): page for page in chapter.pages}
            executor.shutdown()

        for future in as_completed(futures):
            page = futures[future]
            try:
                _ = future.result()
            except Exception as err:
                logger.error(f"Error while downloading {page}: {err}")

    def __make_sync_download(self, chapter: Chapter, downloader: IDownloaderAdapter) -> None:
        for page in chapter.pages:
            self.__make_download_with_queue(page=page)
    
    def __create_queue(self) -> queue.Queue[File]:
        return queue.Queue[File]()

    def execute(self, dto: DTO[DownloadChapterRequestDTO]) -> DownloadChapterResponseDTO:
        try:

            using_threads = dto.data.threaded is not None

            logger.info(f'Downloading chapter with {len(dto.data.chapter.pages)} pages.')
            
            logger.debug(f'Using threads to perform download: {using_threads}.')

            self.queue = self.__create_queue()
            self.downloader = dto.data.downloader

            threads: list[threading.Thread] = []
            queue_threads_num = 1

            logger.debug(f'Spawning {queue_threads_num} threads to handle download queue.')
            for _ in range(queue_threads_num):
                t = threading.Thread(target=self.__handle_saving, args=(self.queue,
                                                                        dto.data.path,
                                                                        dto.data.saver_adapter))
                t.start()
                threads.append(t)

            if dto.data.threaded is not None:
                self.__spawn_workers(chapter=dto.data.chapter, workers=dto.data.threaded.number_of_threads)
            else:
                self.__make_sync_download(chapter=dto.data.chapter, downloader=dto.data.downloader)

            self.queue.join()

            for thread in threads:
                thread.join(timeout=5)

            return DownloadChapterResponseDTO(done=True)
        except Exception as err:
            logger.error(err)
            return DownloadChapterResponseDTO(done=False)



from enum import Enum
import logging
from typing import Optional

class LogMode(Enum):
    DEBUG = logging.DEBUG
    SILENT = -1
    NORMAL = logging.INFO

class CustomFormatter(logging.Formatter):
    grey = "\x1b[38;21m"
    green = "\x1b[32;21m"
    yellow = "\x1b[33;21m"
    red = "\x1b[31;21m"
    bold_red = "\x1b[31;1m"
    reset = "\x1b[0m"

    _format = "%(name)s - [%(levelname)s]: %(message)s (%(filename)s:%(lineno)d)"

    def __init__(self, template: Optional[str] = None):
        
        if template is not None:
            self._format = template

        self.FORMATS = {
            logging.DEBUG: self.grey + self._format + self.reset,
            logging.INFO: self.green + self._format + self.reset,
            logging.WARNING: self.yellow + self._format + self.reset,
            logging.ERROR: self.red + self._format + self.reset,
            logging.CRITICAL: self.bold_red + self._format + self.reset
        }
        

    def format(self, record):
        log_fmt = self.FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt, datefmt="%Y-%m-%d %H:%M:%S")
        return formatter.format(record)

class Logger(logging.Logger):
    def __init__(self, name: str) -> None:
        super().__init__(name=name)
        self._mode = LogMode.SILENT
        self.create_handlers()
        
    def create_handlers(self) -> None:
        debug_formatter = CustomFormatter(template="%(asctime)s - %(name)s - [DEBUG] - %(message)s (in %(funcName)s at %(filename)s:%(lineno)d)")
        info_formatter = CustomFormatter(template="%(name)s - [%(levelname)s]: %(message)s")

        handler = logging.StreamHandler()
        handler.setFormatter(debug_formatter)

        self.addHandler(handler)

        self._debug_formatter = debug_formatter
        self._info_formatter = info_formatter

    @property
    def mode(self) -> LogMode:
        return self._mode

    @mode.setter
    def mode(self, mode: LogMode) -> None:
        if mode == LogMode.SILENT:
            self.disabled = True
            return
        
        if mode == LogMode.DEBUG:
            for handler in self.handlers:
                handler.setFormatter(self._debug_formatter)
        else:
            for handler in self.handlers:
                handler.setFormatter(self._info_formatter)
        
        self.disabled = False
        self.setLevel(mode.value)
        self._mode = mode

logger = Logger('Enma')



import datetime
import json
from unittest.mock import Mock, patch
import sys
import os

import pytest

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '../')))

from enma.application.core.handlers.error import NotFound
from enma.application.use_cases.fetch_chapter_by_symbolic_link import FetchChapterBySymbolicLinkRequestDTO, FetchChapterBySymbolicLinkUseCase
from enma.application.core.interfaces.use_case import DTO
from enma.infra.adapters.repositories.nhentai import CloudFlareConfig, NHentai
from enma.domain.entities.manga import MIME, Author, Chapter, Genre, Image, Manga, SymbolicLink, Title

class TestFetchChapterWithSymbolicLink:

    nhentai = NHentai(config=CloudFlareConfig(user_agent='mocked',
                                              cf_clearance='mocked'))
    sut = FetchChapterBySymbolicLinkUseCase(manga_repository=nhentai)

    mocked_manga = Manga(title=Title(english="[Hikoushiki (CowBow)] Marine Senchou no Yopparai Archive | Marine's Drunken Archives (Houshou Marine) [English] [Watson] [Digital]",
                                     japanese='[飛行式 (矼房)] マリン船長の酔っぱっぱアーカイブ (宝鐘マリン) [英訳] [DL版]',
                                     other="Marine Senchou no Yopparai Archive | Marine's Drunken Archives"),
                         status='completed',
                         url="mocked",
                         id=489504,
                         created_at=datetime.datetime(2024, 1, 7, 0, 3, 25, tzinfo=datetime.timezone.utc),
                         updated_at=datetime.datetime(2024, 1, 7, 0, 3, 25, tzinfo=datetime.timezone.utc),
                         authors=[Author(name='cowbow')],
                         genres=[Genre(name='sweating', id=1590)],
                         thumbnail=Image(uri='https://i.nhentai.net/galleries/2786266/22.jpg', name='21.jpg', width=1280, height=1808, mime=MIME.J),
                         cover=Image(uri='https://i.nhentai.net/galleries/2786266/22.jpg', name='21.jpg', width=1280, height=1808, mime=MIME.J),
                         chapters=[Chapter(id=0, pages=[Image(uri='https://i.nhentai.net/galleries/2786266/1.jpg', name='0.jpg', width=1280, height=1807, mime=MIME.J),
                                                        Image(uri='https://i.nhentai.net/galleries/2786266/2.jpg', name='1.jpg', width=1280, height=1807, mime=MIME.J)])])

    def test_fetch_chapter_by_symbolic_link(self):
        with patch('requests.get') as mock_method:
            mock = Mock()
            mock.status_code = 200
            
            with open('./tests/data/get.json', 'r') as get:
                data = json.loads(get.read())
                mock.json.return_value = data

            mock_method.return_value = mock

            link = SymbolicLink(link='https://nhentai.net/api/gallery/1')
            response = self.sut.execute(dto=DTO(data=FetchChapterBySymbolicLinkRequestDTO(link=link)))

            assert isinstance(response.chapter, Chapter)
            assert response.chapter.pages_count == 14
            assert response.chapter.id == 0
            assert len(response.chapter.pages) == 14

    def test_should_raise_exception_for_broken_link(self):
        with patch('requests.get') as mock_method:
            mock = Mock()
            mock.status_code = 404
            mock.json.return_value = {}
            mock_method.return_value = mock

            link = SymbolicLink(link='https://nhentai.net')
            
            with pytest.raises(NotFound):
                self.sut.execute(dto=DTO(data=FetchChapterBySymbolicLinkRequestDTO(link=link)))
        
    def test_should_return_empty_chapter_for_broken_response(self):
        with patch('requests.get') as mock_method:
            mock = Mock()
            mock.status_code = 200
            mock.json.return_value = {}
            mock_method.return_value = mock

            link = SymbolicLink(link='https://nhentai.net')
            response = self.sut.execute(dto=DTO(data=FetchChapterBySymbolicLinkRequestDTO(link=link)))
            
            assert isinstance(response.chapter, Chapter)
            assert response.chapter.link is None
            assert response.chapter.pages_count == 0
            assert response.chapter.id == 0
            assert len(response.chapter.pages) == 0



from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Generic, Optional, TypeVar

K = TypeVar('K')
V = TypeVar('V')

@dataclass
class DTO(Generic[K]):
    data: K

class IUseCase(ABC, Generic[K, V]):
    @abstractmethod
    def execute(self, dto: Optional[DTO[K]] = None) -> V:
        ...



from typing import Literal, TypedDict, Union


class NHentaiImage(TypedDict):
    t: Union[Literal["j"], Literal["p"], Literal["g"]]
    w: int
    h: int
    
class NHentaiImages(TypedDict):
    pages: list[NHentaiImage]
    cover: NHentaiImage
    thumbnail: NHentaiImage

class Title(TypedDict):
    english: str
    japanese: str
    pretty: str

class Tag(TypedDict):
    id: int
    type: str
    name: str
    url: str
    count: int

class NHentaiResponse(TypedDict):
    id: int
    media_id: str
    title: Title
    images: NHentaiImages
    scanlator: str
    upload_date: int
    tags: list[Tag]
    num_pages: int
    num_favorites: int

class NHentaiPaginateResponse(TypedDict):
    result: NHentaiResponse
    num_pages: int
    per_page: int



import os

from unittest.mock import MagicMock, Mock, patch
import sys

import pytest


sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '../')))

from enma.application.core.handlers.error import Forbidden, NotFound, Unknown
from enma.domain.entities.pagination import Thumb
from enma.infra.adapters.repositories.mangadex import Mangadex
from enma.domain.entities.manga import Author, Chapter, Genre, Image, SymbolicLink

from tests.data.mocked_doujins import mangadex_paginate_mocked

class TestMangadexSourceGetMethod:

    sut = Mangadex()

    def test_success_doujin_retrieve(self):
        
        res = self.sut.get(identifier='65498ee8-3c32-4228-b433-73a4d08f8927', 
                           with_symbolic_links=True)

        assert res is not None
        assert res.id == '65498ee8-3c32-4228-b433-73a4d08f8927'
        assert res.title.english == "Monster Musume no Iru Nichijou"
        assert res.title.japanese == "Monster Musume no Iru Nichijō"
        assert res.title.other != ''
        assert res.url != ''
        
        for genre in res.genres:
            assert isinstance(genre, Genre)
        
        for author in res.authors:
            assert isinstance(author, Author)

        assert isinstance(res.thumbnail, Image)
        assert isinstance(res.cover, Image)

        for chapter in res.chapters:
            assert isinstance(chapter, Chapter)
            assert isinstance(chapter.link, SymbolicLink)

    def test_response_when_it_could_not_get_doujin(self):
        with pytest.raises(NotFound):
            self.sut.get(identifier='manga-kb951984', with_symbolic_links=True)
        
    @patch('requests.get')
    def test_raise_forbidden_in_case_of_403_status_code(self, mock_method: MagicMock):
        mock = Mock()
        mock.status_code = 403
        mock_method.return_value = mock

        with pytest.raises(Forbidden):
            self.sut.get(identifier='manga-kb951984')
            mock_method.assert_called_with(url='https://chapMangadex.com/manga-kb951984', 
                                        headers={'Referer': 'https://chapMangadex.com/'}, 
                                        params={})

    @patch.object(sut, '_Mangadex__list_chapters', autospec=True)
    def test_return_empty_chapters(self, mock_method: MagicMock):
        mock_method.return_value = []

        doujin = self.sut.get(identifier='65498ee8-3c32-4228-b433-73a4d08f8927', 
                              with_symbolic_links=True)
        
        assert doujin is not None

        assert len(doujin.chapters) == 0
        assert doujin.id == '65498ee8-3c32-4228-b433-73a4d08f8927'
        assert doujin.title.english == "Monster Musume no Iru Nichijou"
        assert doujin.title.japanese == "Monster Musume no Iru Nichijō"
        
        for genre in doujin.genres:
            assert isinstance(genre, Genre)
        
        for author in doujin.authors:
            assert isinstance(author, Author)

        assert isinstance(doujin.thumbnail, Image)
        assert isinstance(doujin.cover, Image)
    
    def test_get_with_symbolic_link(self):

            doujin = self.sut.get(identifier='65498ee8-3c32-4228-b433-73a4d08f8927', with_symbolic_links=True)

            assert doujin is not None
            assert isinstance(doujin.chapters[0], Chapter)
            assert doujin.chapters[0].link is not None
            assert doujin.chapters[0].link != ""

class TestMangadexSourcePaginationMethod:
    sut = Mangadex()

    def test_success_pagination(self):
        res = self.sut.paginate(page=2)

        assert res is not None
        assert res.id is not None
        assert res.page == 2
        assert res.total_pages > 0
        assert res.total_results > 0
        assert len(res.results) == 25
        
    @patch('requests.get')
    def test_must_return_empty_search_result(self, mock_method: MagicMock):
        mock = Mock()
        mock.status_code = 200
        mock.json.return_value = mangadex_paginate_mocked
        mock_method.return_value = mock
        
        res = self.sut.paginate(page=2)

        assert res is not None
        assert res.id is not None
        assert res.page == 2
        assert res.total_pages == 0
        assert res.total_results == 0
        assert len(res.results) == 0

    @patch('requests.get')
    def test_response_when_forbidden(self, mock_method: MagicMock):
        mock = Mock()
        mock.status_code = 403
        mock_method.return_value = mock
        
        with pytest.raises(Forbidden):
            self.sut.paginate(page=2)

class TestMangadexSourceSearchMethod:

    sut = Mangadex()

    def test_success_searching(self):

        res = self.sut.search(query='GATE', page=1)

        assert res is not None
        assert res.query == 'GATE'
        assert res.id is not None
        assert res.page == 1
        assert res.total_pages > 0
        assert len(res.results) > 0

        for result in res.results:
            assert isinstance(result, Thumb)
            assert result.cover is not None
            assert isinstance(result.cover, Image)
            assert isinstance(result.cover.width, int)
            assert isinstance(result.cover.height, int)
            assert result.cover.width == 512
            assert result.cover.height == 0
            assert result.title is not None
            assert result.id != 0

    def test_success_searching_using_per_page_param(self):

        res = self.sut.search(query='GATE', page=1, per_page=1)

        assert res is not None
        assert res.query == 'GATE'
        assert res.id is not None
        assert res.page == 1
        assert res.total_pages > 0
        assert len(res.results) == 1

        for result in res.results:
            assert isinstance(result, Thumb)
            assert result.cover is not None
            assert isinstance(result.cover, Image)
            assert isinstance(result.cover.width, int)
            assert isinstance(result.cover.height, int)
            assert result.cover.width == 512
            assert result.cover.height == 0
            assert result.title is not None
            assert result.id != 0
        
    @patch('requests.get')
    def test_response_when_forbidden(self, mock_method: MagicMock):
        mock = Mock()
        mock.status_code = 403
        mock_method.return_value = mock

        with pytest.raises(Forbidden):
            self.sut.search(query='Monster Musume no Iru Nichijou', page=1)
    
    @patch('requests.get')
    def test_response_when_not_found(self, mock_method: MagicMock):
        mock = Mock()
        mock.status_code = 404
        mock_method.return_value = mock

        with pytest.raises(NotFound):
            self.sut.search(query='Monster Musume no Iru Nichijou', page=1)
    
    @patch('requests.get')
    def test_response_when_unknown(self, mock_method: MagicMock):
        mock = Mock()
        mock.status_code = 500
        mock_method.return_value = mock

        with pytest.raises(Unknown):
            self.sut.search(query='Monster Musume no Iru Nichijou', page=1)



import sys
import os

import pytest

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '../')))

from enma.application.core.interfaces.manga_repository import IMangaRepository
from enma.infra.adapters.repositories.mangadex import Mangadex
from enma.infra.entrypoints.lib import SourceManager, Sources
from enma.application.core.handlers.error import InstanceError, SourceNotAvailable

class TestSourceManager:

    sut = SourceManager()

    @pytest.fixture(autouse=True)
    def run_around_tests(self):
        yield
        self.sut.clear_sources()

    def test_add_source_successfully(self):
        self.sut.add_source(source_name=Sources.MANGADEX, source=Mangadex())
        assert self.sut.get_source(Sources.MANGADEX) is not None

    def test_get_source_should_raise_exception_if_source_doesnt_exists(self):
        with pytest.raises(SourceNotAvailable):
            self.sut.get_source(Sources.NHENTAI)

    def test_get_source_successfully_with_string_parameter(self):
        self.sut.add_source(source_name=Sources.MANGADEX, source=Mangadex())
        assert isinstance(self.sut.get_source('mangadex'), IMangaRepository)

    def test_clear_sources_successfully(self):
        self.sut.add_source(source_name=Sources.MANGADEX, source=Mangadex())
        self.sut.clear_sources()
        with pytest.raises(SourceNotAvailable):
            self.sut.get_source(Sources.MANGADEX)

    def test_remove_source_successfully(self):
        self.sut.add_source(source_name=Sources.MANGADEX, source=Mangadex())
        assert self.sut.remove_source(Sources.MANGADEX) is True
        with pytest.raises(SourceNotAvailable):
            self.sut.get_source(Sources.MANGADEX)

    def test_remove_source_should_return_false_if_source_doesnt_exists(self):
        assert self.sut.remove_source(Sources.MANGADEX) is False

    def test_set_source_successfully(self):
        self.sut.add_source(source_name=Sources.MANGADEX, source=Mangadex())
        self.sut.set_source(Sources.MANGADEX)
        assert self.sut.source is not None

    def test_should_not_be_able_to_add_an_source_that_is_not_imanga_repository(self):
        with pytest.raises(InstanceError):
            self.sut.add_source(source_name=Sources.MANGADEX, source=object()) # type: ignore



from enma.application.core.interfaces.saver_adapter import File, ISaverAdapter
from enma.application.core.utils.logger import logger

class GoogleDriveStorage(ISaverAdapter):

    def __init__(
            self, 
            credentials_path: str,
            root_shared_folder: str):

        try:
            from googleapiclient.discovery import build, HttpError
            from googleapiclient.http import MediaIoBaseUpload
            from google.oauth2.service_account import Credentials

            self.MediaIoBaseUpload = MediaIoBaseUpload
            self.HttpError = HttpError
            self.build = build
        except ImportError as e:
            raise ImportError(
                "The dependencies for Google Drive are not installed. "
                "Please install them using 'pip install enma[google_drive]'."
            ) from e
            
        self.credentials = Credentials.from_service_account_file(credentials_path)
        self.service = build('drive', 'v3', credentials=self.credentials)
        self.root_shared_folder = root_shared_folder

    def save(self, path: str, file: File) -> bool:
        try:
            folder_id = self._get_or_create_folder(path)

            file_metadata = {
                'name': file.name,
                'parents': [folder_id]
            }

            logger.debug(f'Uploading image to google drive with name: {file.name} and parent folder: {folder_id}')

            media = self.MediaIoBaseUpload(file.data, mimetype='application/octet-stream')

            self.service.files().create(
                body=file_metadata,
                media_body=media,
                fields='id'
            ).execute()
            return True
        except self.HttpError as e:
            logger.error(f'A HTTP error ocurred while trying to upload image to google drive: {e}')
            return False
        except Exception as e:
            logger.error(f'An unknown error ocurred while trying to upload image to google drive: {e}')
            return False

    def _get_or_create_folder(self, path: str) -> str:
        folder_names = path.strip('/').split('/')
        parent_id = self.root_shared_folder

        for folder_name in folder_names:
            query = (
                f"name = '{folder_name}' and "
                f"mimeType = 'application/vnd.google-apps.folder' and "
                f"'{parent_id}' in parents and "
                f"trashed = false"
            )

            logger.debug(f'Querying folder with query: {query}')

            results = self.service.files().list(
                q=query,
                spaces='drive',
                fields='files(id, name)'
            ).execute()
            items = results.get('files', [])

            if items:
                folder_id = items[0]['id']
            else:
                folder_metadata = {
                    'name': folder_name,
                    'mimeType': 'application/vnd.google-apps.folder',
                    'parents': [parent_id]
                }

                logger.debug(f'Creating folder with metadata: {folder_metadata}')

                folder = self.service.files().create(
                    body=folder_metadata,
                    fields='id'
                ).execute()
                folder_id = folder.get('id')

            parent_id = folder_id

        return parent_id


import datetime
import json
from unittest.mock import MagicMock, Mock, patch
import sys
import os

import pytest
from pydantic import ValidationError

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '../')))

from enma.application.core.handlers.error import Forbidden, NotFound
from enma.infra.core.interfaces.nhentai_response import NHentaiResponse
from enma.application.use_cases.get_manga import GetMangaRequestDTO, GetMangaUseCase
from enma.application.core.interfaces.use_case import DTO
from enma.infra.adapters.repositories.nhentai import CloudFlareConfig, NHentai
from enma.domain.entities.manga import MIME, Author, Chapter, Genre, Image, Manga, Title

class TestNHentaiGetDoujin:

    nhentai = NHentai(config=CloudFlareConfig(user_agent='mocked',
                                              cf_clearance='mocked'))
    sut = GetMangaUseCase(manga_repository=nhentai)

    mocked_manga = Manga(title=Title(english="[Hikoushiki (CowBow)] Marine Senchou no Yopparai Archive | Marine's Drunken Archives (Houshou Marine) [English] [Watson] [Digital]",
                                     japanese='[飛行式 (矼房)] マリン船長の酔っぱっぱアーカイブ (宝鐘マリン) [英訳] [DL版]',
                                     other="Marine Senchou no Yopparai Archive | Marine's Drunken Archives"),
                         status='completed',
                         url='mocked',
                         id=489504,
                         created_at=datetime.datetime(2024, 1, 7, 0, 3, 25, tzinfo=datetime.timezone.utc),
                         updated_at=datetime.datetime(2024, 1, 7, 0, 3, 25, tzinfo=datetime.timezone.utc),
                         authors=[Author(name='cowbow')],
                         genres=[Genre(name='sweating', id=1590)],
                         thumbnail=Image(uri='https://i.nhentai.net/galleries/2786266/22.jpg', name='21.jpg', width=1280, height=1808, mime=MIME.J),
                         cover=Image(uri='https://i.nhentai.net/galleries/2786266/22.jpg', name='21.jpg', width=1280, height=1808, mime=MIME.J),
                         chapters=[Chapter(id=0, pages=[Image(uri='https://i.nhentai.net/galleries/2786266/1.jpg', name='0.jpg', width=1280, height=1807, mime=MIME.J),
                                                        Image(uri='https://i.nhentai.net/galleries/2786266/2.jpg', name='1.jpg', width=1280, height=1807, mime=MIME.J)])])

    def test_success_doujin_retrieve(self):
        with patch('enma.infra.adapters.repositories.nhentai.NHentai.get') as mock_method:
            mock_method.return_value = self.mocked_manga
            res = self.sut.execute(dto=DTO(data=GetMangaRequestDTO(identifier='489504')))

            assert res.found == True
            assert res.manga is not None
            assert res.manga.id == 489504
            assert res.manga.title.english == "[Hikoushiki (CowBow)] Marine Senchou no Yopparai Archive | Marine's Drunken Archives (Houshou Marine) [English] [Watson] [Digital]"
            assert res.manga.title.japanese == "[飛行式 (矼房)] マリン船長の酔っぱっぱアーカイブ (宝鐘マリン) [英訳] [DL版]"
            assert res.manga.title.other == "Marine Senchou no Yopparai Archive | Marine's Drunken Archives"
            
            for genre in res.manga.genres:
                assert isinstance(genre, Genre)
            
            for author in res.manga.authors:
                assert isinstance(author, Author)

            assert isinstance(res.manga.thumbnail, Image)
            assert isinstance(res.manga.cover, Image)

            for chapter in res.manga.chapters:
                assert isinstance(chapter, Chapter)
            
            mock_method.assert_called_with(identifier='489504', with_symbolic_links=False)

    def test_must_return_other_titles_as_none_if_doesnt_exists(self):
        with patch('requests.get') as mock_method:
            mock = Mock()
            mock.status_code = 200
            
            with open('./tests/data/get.json', 'r') as get:
                data: NHentaiResponse = json.loads(get.read())
                
                data['title']['japanese'] = None # type: ignore
                data['title']['pretty'] = None # type: ignore

                mock.json.return_value = data
            
            mock_method.return_value = mock
            
            res = self.sut.execute(dto=DTO(data=GetMangaRequestDTO(identifier='489504')))

            assert res.found == True
            assert res.manga is not None
            assert res.manga.id == 1
            assert res.manga.title.english == "(C71) [Arisan-Antenna (Koari)] Eat The Rich! (Sukatto Golf Pangya)"
            assert res.manga.title.japanese == None
            assert res.manga.title.other == None
        
    def test_response_when_it_could_not_get_doujin(self):
        with patch('enma.infra.adapters.repositories.nhentai.NHentai.get', side_effect=NotFound("Could not find the manga")) as mock_method:
            doujin = self.sut.execute(dto=DTO(data=GetMangaRequestDTO(identifier='1')))
            assert doujin.found == False
            assert doujin.manga is None
    
    def test_raise_forbidden_in_case_of_403(self):
        with patch('requests.get') as mock_method:
            mock = Mock()
            mock.status_code = 403
            mock_method.return_value = mock

            with pytest.raises(Forbidden):
                self.sut.execute(dto=DTO(data=GetMangaRequestDTO(identifier='1')))
                mock_method.assert_called_with(url=f'https://nhentai.net/api//gallery/1',
                                            headers={'User-Agent': 'mocked'},
                                            params={},
                                            cookies={'cf_clearance': 'mocked'})

    def test_return_empty_chapters(self):
        with patch('requests.get') as mock_method:
            mock = Mock()
            mock.status_code = 200

            with open('./tests/data/get.json', 'r') as get:
                data = json.loads(get.read())
                data['images']['pages'] = []
                mock.json.return_value = data

            mock_method.return_value = mock

            doujin = self.sut.execute(dto=DTO(data=GetMangaRequestDTO(identifier='1')))

            assert doujin.found == True
            assert doujin.manga is not None
            assert isinstance(doujin.manga.chapters[0], Chapter)
            assert len(doujin.manga.chapters[0].pages) == 0
            assert doujin.manga.id == 1 

    def test_return_right_mime(self):
        with patch('requests.get') as mock_method:
            mock = Mock()
            mock.status_code = 200

            with open('./tests/data/get.json', 'r') as get:
                data = json.loads(get.read())
                mock.json.return_value = data

            mock_method.return_value = mock

            doujin = self.sut.execute(dto=DTO(data=GetMangaRequestDTO(identifier='2')))

            assert doujin.found == True
            assert doujin.manga is not None
            assert isinstance(doujin.manga.chapters[0], Chapter)
            assert doujin.manga.chapters[0].pages[0].mime.value == 'jpg'
            assert data['images']['pages'][0]['t'] == 'j'
            assert doujin.manga.cover is not None
            assert doujin.manga.cover.mime.value == 'png'
            assert data['images']['cover']['t'] == 'p'
            assert doujin.manga.chapters[0].link is None

    def test_get_with_symbolic_link(self):
        with patch('requests.get') as mock_method:
            mock = Mock()
            mock.status_code = 200

            with open('./tests/data/get.json', 'r') as get:
                data = json.loads(get.read())
                mock.json.return_value = data

            mock_method.return_value = mock

            doujin = self.sut.execute(dto=DTO(data=GetMangaRequestDTO(identifier='420719', with_symbolic_links=True)))

            assert doujin.found == True
            assert doujin.manga is not None
            assert isinstance(doujin.manga.chapters[0], Chapter)
            assert doujin.manga.chapters[0].link is not None
            assert doujin.manga.chapters[0].link != ""

    def test_language_must_be_present(self):
        with patch('requests.get') as mock_method:
            mock = Mock()
            mock.status_code = 200

            with open('./tests/data/get.json', 'r') as get:
                data = json.loads(get.read())
                mock.json.return_value = data

            mock_method.return_value = mock

            doujin = self.sut.execute(dto=DTO(data=GetMangaRequestDTO(identifier='420719', with_symbolic_links=True)))

            assert doujin.found == True
            assert doujin.manga is not None
            assert doujin.manga.language is not None
            assert doujin.manga.language == 'japanese'

    def test_images_mime_types_must_be_correct(self):
        with patch('requests.get') as mock_method:
            mock = Mock()
            mock.status_code = 200
            
            with open('./tests/data/get.json', 'r') as get:
                data: NHentaiResponse = json.loads(get.read())
                mock.json.return_value = data

            mock_method.return_value = mock

            doujin = self.sut.execute(dto=DTO(data=GetMangaRequestDTO(identifier='420719', with_symbolic_links=True)))

            assert doujin.found == True
            assert doujin.manga is not None
            
            cover_mime = data['images']['cover']['t']
            thumb_mime = data['images']['thumbnail']['t']

            assert doujin.manga.thumbnail is not None
            assert doujin.manga.cover is not None
            assert cover_mime.upper() == 'P'
            assert doujin.manga.cover.mime.value == 'png'
            assert thumb_mime.upper() == 'J'
            assert doujin.manga.thumbnail.mime.value == 'jpg'

    @patch('enma.application.use_cases.get_manga.GetMangaUseCase.execute')
    def test_symbolic_links_must_be_disabled_by_default(self, use_case_mock: MagicMock):
        self.sut.execute(dto=DTO(data=GetMangaRequestDTO(identifier='420719')))
        use_case_mock.assert_called_with(dto=DTO(data=GetMangaRequestDTO(identifier='420719', with_symbolic_links=False)))

    def test_must_raise_an_exception_case_user_has_provided_wrong_data_type(self):
        with pytest.raises(ValidationError) as _:
            self.sut.execute(dto=DTO(data=GetMangaRequestDTO(identifier='420719', with_symbolic_links='nao'))) # type: ignore



from unittest.mock import MagicMock, Mock, patch
import sys
import os

import pytest

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '../')))

from enma.domain.entities.pagination import Thumb
from enma.infra.adapters.repositories.manganato import Manganato
from enma.domain.entities.manga import MIME, Author, Chapter, Genre, Image, SymbolicLink
from tests.data.mocked_doujins import manganato_manga_page_empty_chapters_mocked, manganato_manga_page_empty_pagination_mocked
from enma._version import __version__

class TestManganatoSourceGetMethod:

    sut = Manganato()

    def test_success_doujin_retrieve(self):

        res = self.sut.get(identifier='manga-kb951984', with_symbolic_links=True)

        assert res is not None
        assert res.id == 'manga-kb951984'
        assert res.title.english == "Monster Musume No Iru Nichijou"
        assert res.title.japanese == "モンスター娘のいる日常"
        assert res.title.other == "魔物娘的同居日常"
        assert res.url != ''
        
        for genre in res.genres:
            assert isinstance(genre, Genre)
        
        for author in res.authors:
            assert isinstance(author, Author)

        assert isinstance(res.thumbnail, Image)
        assert isinstance(res.cover, Image)

        for chapter in res.chapters:
            assert isinstance(chapter, Chapter)
            assert isinstance(chapter.link, SymbolicLink)

    def test_response_when_it_could_not_get_doujin(self):
        doujin = self.sut.get(identifier='manga-kb951984ss', with_symbolic_links=True)
        
        assert doujin is None
    
    @patch('requests.get')
    def test_return_none_when_not_receive_200_status_code(self, mock_method: MagicMock):
            mock = Mock()
            mock.status_code = 403
            mock_method.return_value = mock

            doujin = self.sut.get(identifier='manga-kb951984')

            assert doujin is None
            mock_method.assert_called_with(url='https://chapmanganato.com/manga-kb951984', 
                                           headers={'Referer': 'https://chapmanganato.com/',
                                                    'User-Agent': f'Enma/{__version__}'}, 
                                           params={})
    
    @patch('requests.get')
    def test_return_empty_chapters(self, mock_method: MagicMock):
        mock = Mock()
        mock.status_code = 200
        mock.text = manganato_manga_page_empty_chapters_mocked

        mock_method.return_value = mock

        doujin = self.sut.get(identifier='manga-kb951984')

        assert doujin is not None
        assert len(doujin.chapters) == 0
        assert doujin.chapters_count == 0
        assert doujin.id == 'manga-kb951984'
        assert doujin.title.english == "Monster Musume No Iru Nichijou"
        assert doujin.title.japanese == "モンスター娘のいる日常"
        assert doujin.title.other == "魔物娘的同居日常"
        for genre in doujin.genres:
            assert isinstance(genre, Genre)
        
        for author in doujin.authors:
            assert isinstance(author, Author)

        assert isinstance(doujin.thumbnail, Image)
        assert isinstance(doujin.cover, Image)
    
    def test_get_with_symbolic_link(self):

            doujin = self.sut.get(identifier='manga-kb951984', with_symbolic_links=True)

            assert doujin is not None
            assert isinstance(doujin.chapters[0], Chapter)
            assert doujin.chapters[0].link is not None
            assert doujin.chapters[0].link != ""

class TestManganatoSourcePaginationMethod:
    sut = Manganato()

    def test_success_searching(self):
        res = self.sut.paginate(page=2)

        assert res is not None
        assert res.id is not None
        assert res.page == 2
        assert res.total_pages > 0
        assert res.total_results > 0
        assert len(res.results) == 24
        
    @patch('requests.get')
    def test_must_return_empty_search_result(self, mock_method: MagicMock):
        mock = Mock()
        mock.status_code = 200

        mock.text = manganato_manga_page_empty_pagination_mocked

        mock_method.return_value = mock
        
        res = self.sut.paginate(page=2)

        assert res is not None
        assert res.id is not None
        assert res.page == 2
        assert res.total_pages > 0
        assert res.total_results > 0
        assert len(res.results) == 0

    @patch('requests.get')
    def test_response_when_forbidden(self, mock_method: MagicMock):
        mock = Mock()
        mock.status_code = 403
        mock_method.return_value = mock
        
        res = self.sut.paginate(page=2)
        
        assert res is not None
        assert res.id is not None
        assert res.page == 2
        assert res.total_pages == 0
        assert res.total_results == 0
        assert len(res.results) == 0

class TestManganatoSourceSearchMethod:

    sut = Manganato()

    def test_success_searching(self):

        res = self.sut.search(query='GATE', page=1)

        assert res is not None
        assert res.query == 'GATE'
        assert res.id is not None
        assert res.page == 1
        assert res.total_pages == 5
        assert len(res.results) == 20

        for result in res.results:
            assert isinstance(result, Thumb)
            assert result.cover is not None
            assert isinstance(result.cover, Image)
            assert isinstance(result.cover.width, int)
            assert isinstance(result.cover.height, int)
            assert result.cover.width == 0
            assert result.cover.height == 0
            assert result.title is not None
            assert result.id != 0
        
    @patch('requests.get')
    def test_response_when_forbidden(self, mock_method: MagicMock):
        mock = Mock()
        mock.status_code = 403
        mock_method.return_value = mock

        search = self.sut.search(query='Monster Musume no Iru Nichijou', page=1)
        
        assert search is not None
        assert search.query == 'Monster Musume no Iru Nichijou'
        assert search.id is not None
        assert search.page == 1
        assert search.total_pages == 0
        assert len(search.results) == 0



import sys
import os

import pytest

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '../')))

from enma.domain.entities.manga import Image
from enma.domain.entities.pagination import Thumb
from enma.domain.entities.search_result import SearchResult

class TestSearchResultEntity:

    def test_default_values(self):
        sut = SearchResult(query='teste',
                           page=1,
                           results=[])
        assert sut.results == []
        assert sut.total_pages == 0
        assert sut.total_pages == 0

    def test_set_results_at_instantiation_shouldnt_effect_counters(self):

        thumbs = [Thumb(id='1', url='mocked', title='teste', cover=Image(uri='teste')),
                  Thumb(id='2', url='mocked', title='teste', cover=Image(uri='teste')),
                  Thumb(id='3', url='mocked', title='teste', cover=Image(uri='teste'))]

        sut = SearchResult(query='teste',
                           page=1,
                           results=thumbs)
        
        assert sut.results == thumbs
        assert sut.page == 1
        assert sut.total_results == 0
        assert sut.total_pages == 0
        assert sut.results[0].url == 'mocked'

    def test_set_results_at_instantiation_shouldnt_effect_counters_after(self):

        thumbs = [Thumb(id='1', url='mocked', title='teste', cover=Image(uri='teste')),
                  Thumb(id='2', url='mocked', title='teste', cover=Image(uri='teste')),
                  Thumb(id='3', url='mocked', title='teste', cover=Image(uri='teste'))]

        sut = SearchResult(query='teste',
                           page=1,
                           total_pages=100,
                           total_results=300,
                           results=thumbs)
        
        assert sut.results == thumbs
        assert sut.page == 1
        assert sut.total_results == 300
        assert sut.total_pages == 100
